#!/usr/bin/env python3
"""
Create HTML report from QC analysis results

AUTHORS
----
Mike Tyszka, Ph.D., Caltech Brain Imaging Center

DATES
----
2013-09-25 JMT From scratch
2013-10-23 JMT Add com external call
2013-10-24 JMT Move stats calcs to new stats.py
2019-05-28 JMT Recode as a nipype interface class
2019-05-29 JMT Expand to multiple pages

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import json
import shutil
import datetime as dt

from reportlab.lib.enums import TA_JUSTIFY
from reportlab.lib.pagesizes import letter
from reportlab.platypus import (SimpleDocTemplate,
                                Paragraph,
                                Spacer,
                                Image,
                                Table,
                                TableStyle,
                                PageBreak)
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch


class ReportPDF:

    def __init__(self, fnames, meta, metrics):

        self._fnames = fnames
        self._meta = meta
        self._metrics = metrics
        self._tmp_report_pdf = os.path.join(fnames['WorkDir'], 'report.pdf')

        # Contents - list of flowables to be built into a document
        self._contents = []

        # Add a justified paragraph style
        self._pstyles = getSampleStyleSheet()
        self._pstyles.add(ParagraphStyle(name='Justify', alignment=TA_JUSTIFY))

        self._init_pdf()
        self._add_summary()
        self._add_roi_timeseries()
        self._add_motion_timeseries()
        self._add_sections()
        self._add_demeaned_ts()

        self._doc.build(self._contents)
        self._save_report()

    def _init_pdf(self):

        # Create a new PDF document
        self._doc = SimpleDocTemplate(self._tmp_report_pdf,
                                      pagesize=letter,
                                      rightMargin=0.5 * inch,
                                      leftMargin=0.5 * inch,
                                      topMargin=0.5 * inch,
                                      bottomMargin=0.5 * inch)

    def _add_summary(self):

        ptext = '<font size=24>CBIC Quality Control Report</font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.5 * inch))

        timestamp = dt.datetime.now().strftime('%Y-%m-%d at %H:%M:%S')
        ptext = '<font size=12>Generated by CBICQC on {}</font>'.format(timestamp)
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        #
        # Session information
        #

        ptext = '<font size=14><b>Session Information</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        demo = [['Subject', self._meta['Subject']],
                ['Session', self._meta['Session']],
                ['Scan Time', self._meta['AcquisitionTime']],
                ['Software Version', self._meta['SoftwareVersions']],
                ['Pulse Sequence', self._meta['SequenceName']],
                ['Coil Name', self._meta['ReceiveCoilName']],
                ['', '']
                ]

        image = [['Repetition Time', '{} ms'.format(self._meta['RepetitionTime'] * 1e3)],
                 ['Echo Time', '{} ms'.format(self._meta['EchoTime'] * 1e3)],
                 ['Voxel Size', '{} mm'.format(self._meta['VoxelSize'])],
                 ['Matrix Size', self._meta['MatrixSize']],
                 ['EPI Echo Spacing', '{0:.0f} us'.format(self._meta['EffectiveEchoSpacing'] * 1e6)],
                 ['RO Bandwidth', '{} Hz/pix'.format(self._meta['PixelBandwidth'])],
                 ['PE Bandwidth', '{} Hz/pix'.format(self._meta['BandwidthPerPixelPhaseEncode'])]]

        demo_table = Table(demo, hAlign='LEFT')
        image_table = Table(image, hAlign='LEFT')

        self._contents.append(Table([[demo_table, image_table]]))
        self._contents.append(Spacer(1, 0.25 * inch))

        #
        # QC metrics
        #

        ptext = '<font size=14><b>Quality Metrics</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        signal_metrics = [['Mean Signal', '{:.1f}'.format(self._metrics['SignalMean'])],
                          ['SNR', '{:.1f}'.format(self._metrics['SNR'])],
                          ['SFNR', '{:.1f}'.format(self._metrics['SFNR'])],
                          ['SArtR', '{:.1f}'.format(self._metrics['SArtR'])],
                          ['Drift', '{:.3f} %/TR'.format(self._metrics['Drift'])],
                          ['Warmup Amplitude', '{:.3f} %'.format(self._metrics['WarmupAmp'])],
                          ['Warmup Time Constant', '{:.1f} TRs'.format(self._metrics['WarmupTime'])]
                          ]

        ptext = '<font size=11><b>Main Signal</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        signal_table = Table(signal_metrics, hAlign='LEFT')
        self._contents.append(signal_table)
        self._contents.append(Spacer(1, 0.25 * inch))

        noise_metrics = [['Noise Sigma', '{:.1f}'.format(self._metrics['NoiseSigma'])],
                         ['Noise Floor', '{:.1f}'.format(self._metrics['NoiseFloor'])],
                         ['Signal Spikes', '{}'.format(self._metrics['SignalSpikes'])],
                         ['Nyquist Ghost Spikes', '{}'.format(self._metrics['NyquistSpikes'])],
                         ['Air Spikes', '{}'.format(self._metrics['AirSpikes'])],
                         ]

        ptext = '<font size=11><b>Noise and Spiking</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        noise_table = Table(noise_metrics, hAlign='LEFT')
        self._contents.append(noise_table)
        self._contents.append(Spacer(1, 0.25 * inch))

    def _add_roi_timeseries(self):

        # Page break
        self._contents.append(PageBreak())

        #
        # ROI Spatial Mean Timeseries
        #

        ptext = '<font size=14><b>ROI Spatial Mean Timeseries</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = """
        <font size=11>
        These three graphs show the spatial mean signal intensity within the main signal, Nyquist ghost and air
        space regions of interest (ROIs). The blue line is the raw spatial mean intensity and the orange line
        has been detrended with an exponential + linear model. 
        </font>
        """
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        roi_ts_img = Image(self._fnames['ROITimeseries'], 7.0 * inch, 3.5 * inch, hAlign='LEFT')
        self._contents.append(roi_ts_img)

        self._contents.append(Spacer(1, 0.5 * inch))

        #
        # ROI Spatial Mean Power Spectrum
        #

        ptext = '<font size=14><b>ROI Power Spectra</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = """
        <font size=11>
        Power spectrum of the spatial mean signal in each ROI. The dB scale is relative to maximum spectral power.
        </font>
        """
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        roi_ps_img = Image(self._fnames['ROIPowerspec'], 7.0 * inch, 3.5 * inch, hAlign='LEFT')
        self._contents.append(roi_ps_img)

        self._contents.append(Spacer(1, 0.5 * inch))


    def _add_motion_timeseries(self):

        # Page break
        self._contents.append(PageBreak())

        #
        # Motion Parameter Timeseries
        #

        ptext = '<font size=14><b>Motion Parameter Timeseries</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        # Main signal QC specific text
        ptext = """
        <font size=11>
        Displacement and rotation parameter timecourses required
        to register the center of mass at a given time to the center of mass of the first image. 
        </font>
        """
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        mo_ts_img = Image(self._fnames['MoparTimeseries'], 7.0 * inch, 3.5 * inch, hAlign='LEFT')
        self._contents.append(mo_ts_img)

        self._contents.append(Spacer(1, 0.5 * inch))

        #
        # Motion Power Spectra
        #

        ptext = '<font size=14><b>Motion Power Spectra</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = """
        <font size=11>
        Power spectrum of the absolute displacement and total rotation timecourses.
        The dB scale is relative to maximum spectral power.
        </font>
        """
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        mo_ts_img = Image(self._fnames['MoparPowerspec'], 7.0 * inch, 3.5 * inch, hAlign='LEFT')
        self._contents.append(mo_ts_img)

        self._contents.append(Spacer(1, 0.5 * inch))

    def _add_sections(self):

        # Page break
        self._contents.append(PageBreak())

        ptext = '<font size=14><b>Image Sections</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        self._add_section('Temporal Mean', self._fnames['TMeanMontage'])
        self._add_section('Temporal Standard Deviation', self._fnames['TSDMontage'])
        self._add_section('Regions of Interest', self._fnames['ROIsMontage'])

    def _add_section(self, title, img_fname):

        ptext = '<font size=11><b>{}</b></font>'.format(title)
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        tmean_montage_img = Image(img_fname, 7.0 * inch, 2.4 * inch, hAlign='LEFT')

        self._contents.append(tmean_montage_img)
        self._contents.append(Spacer(1, 0.25 * inch))

    def _add_demeaned_ts(self):

        # Page break
        self._contents.append(PageBreak())

        ptext = '<font size=14><b>Demeaned Signal</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = """
        <font size=11>
        Demeaned signal timecourses for a subsample of voxels from each ROI.
        </font>
        """
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        residuals_img = Image(self._fnames['ROIDemeanedTS'], 7.0 * inch, 9.0 * inch, hAlign='LEFT')
        self._contents.append(residuals_img)

    def _save_report(self):

        # Copy report PDF to derivatives
        shutil.copyfile(self._tmp_report_pdf, self._fnames['ReportPDF'])

        # Save metrics in derivatives as JSON file
        with open(self._fnames['ReportJSON'], 'w') as fd:
            json.dump(self._metrics, fd, sort_keys=True, indent=4)





#!/usr/bin/env python3
"""
Create summary PDF report

AUTHORS
----
Mike Tyszka, Ph.D., Caltech Brain Imaging Center

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import shutil
import tempfile
from datetime import datetime

import matplotlib.pyplot as plt
from pandas.plotting import register_matplotlib_converters
import numpy as np
from reportlab.lib.enums import TA_JUSTIFY
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
from reportlab.platypus import (SimpleDocTemplate,
                                Paragraph,
                                Spacer,
                                Image,
                                Table)

from .graphics import trend_subplot


class SummaryPDF:

    def __init__(self, report_dir, metrics):
        """
        Construct summary report PDF

        :param report_dir: str, report output directory in derivatives
        :param metrics: list, session metric dictionaries including metadata
        """

        # For datetime axis labeling without warnings
        register_matplotlib_converters()

        self._report_dir = report_dir
        self._metrics = metrics
        self._subject = metrics[0]['Subject']
        self._summary_pdf = os.path.join(report_dir, '{}_summary.pdf'.format(self._subject))

        # Create working directory for images
        self._work_dir = tempfile.mkdtemp()

        # Contents - list of flowables to be built into a document
        self._contents = []

        # Add a justified paragraph style
        self._pstyles = getSampleStyleSheet()
        self._pstyles.add(ParagraphStyle(name='Justify', alignment=TA_JUSTIFY))

        self._init_pdf()
        self._add_coverpage()
        self._add_metric_trends()

        self._doc.build(self._contents)

        # Delete working directory
        shutil.rmtree(self._work_dir)

    def _add_metric_trends(self):

        metric_list = [
            'SNR',
            'SFNR',
            'NoiseFloor',
            'Drift',
            'NyquistSpikes',
            'AirSpikes'
        ]

        trend_png_fname = self._plot_trends(metric_list)

        ptext = '<font size=14><b>Session Metric Trends</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        trends_img = Image(trend_png_fname, 6.0 * inch, 6.0 * inch, hAlign='LEFT')
        self._contents.append(trends_img)

    def _init_pdf(self):

        # Create a new PDF document
        self._doc = SimpleDocTemplate(self._summary_pdf,
                                      pagesize=letter,
                                      rightMargin=0.5 * inch,
                                      leftMargin=0.5 * inch,
                                      topMargin=0.5 * inch,
                                      bottomMargin=0.5 * inch)

    def _add_coverpage(self):

        ptext = '<font size=24>CBIC Quality Control Summary</font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.5 * inch))

        timestamp = datetime.now().strftime('%Y-%m-%d at %H:%M:%S')
        ptext = '<font size=12>Generated by CBICQC on {}</font>'.format(timestamp)
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.25 * inch))

        ptext = '<font size=14><b>Subject Details</b></font>'
        self._contents.append(Paragraph(ptext, self._pstyles['Justify']))
        self._contents.append(Spacer(1, 0.1 * inch))

        # First session metadata - assume identical for all sessions
        m = self._metrics[0]

        meta = [['Subject', self._subject],
                ['Scanner', m['StationName'] + ' ' + m['DeviceSerialNumber']],
                ['Software Version', m['SoftwareVersions']],
                ['Coil Name', m['ReceiveCoilName']]]

        self._contents.append(Table(meta, hAlign='LEFT'))
        self._contents.append(Spacer(1, 0.25 * inch))

    def _plot_trends(self, metric_list):

        nm = len(metric_list)

        t = []
        mm = []

        for mt in self._metrics:

            if 'AcquisitionDateTime' in mt.keys():
                dt = datetime.strptime(mt['AcquisitionDateTime'], '%Y-%m-%dT%H:%M:%S.%f')
                t.append(dt)
                mm.append([mt[m_name] for m_name in metric_list])
            else:
                print('* Acquisition time unknown - skipping')

        t, mm = np.array(t), np.array(mm)

        png_fname = os.path.join(self._work_dir, 'metric_trends.png')

        plt.subplots(nm, 1, figsize=(7, 7))

        for mc, m_name in enumerate(metric_list):

            plt.subplot(nm, 1, mc+1)
            trend_subplot(m_name, t, mm[:, mc])

        # Add time axis label to final subplot
        plt.xlabel('Date')

        plt.subplots_adjust(bottom=0.0, top=0.9, left=0.0, right=1.0)

        # Remove excess space
        plt.tight_layout()

        # Save plot to file
        plt.savefig(png_fname, dpi=300)

        return png_fname



# !/usr/bin/env python
"""
CBICQC timeseries analysis node

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-05-22 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np
from scipy.optimize import least_squares
import nibabel as nb


def temporal_mean_sd(qc_moco_nii):

    # Temporal mean of 4D timeseries
    tmean = np.mean(qc_moco_nii.get_data(), axis=3)
    tsd = np.std(qc_moco_nii.get_data(), axis=3)
    tsfnr = tmean / (tsd + np.finfo(float).eps)

    tmean_nii = nb.Nifti1Image(tmean, qc_moco_nii.affine)
    tsd_nii = nb.Nifti1Image(tsd, qc_moco_nii.affine)
    tsfnr_nii = nb.Nifti1Image(tsfnr, qc_moco_nii.affine)

    return tmean_nii, tsd_nii, tsfnr_nii


def extract_timeseries(qc_moco_nii, rois_nii):

    rois = rois_nii.get_data()
    s = qc_moco_nii.get_data()

    # Number of time points
    nt = s.shape[3]

    # ROI label indices
    # 0 : unassigned
    # 1 : air space
    # 2 : Nyquist ghost
    # 3 : signal

    labels = [1, 2, 3]
    nl = len(labels)

    s_mean_t = np.zeros([nl, nt])

    for lc in range(0, nl):
        mask = np.array(rois == labels[lc])
        for tc in range(0, nt):
            s_t = s[:, :, :, tc]
            s_mean_t[lc, tc] = np.mean(s_t[mask])

    return s_mean_t


def detrend_timeseries(s_mean_t):
    """
    :param s_mean_t: spatial mean ROI timeseries
    :return:
    """

    nl, nt = s_mean_t.shape

    # Time vector
    t = np.arange(0, nt)

    s_detrend_t = np.zeros_like(s_mean_t)

    fit_results = []

    for lc in range(0, nl):

        s_t = s_mean_t[lc, :]

        s_min, s_max, s_mean = np.min(s_t), np.max(s_t), np.mean(s_t)
        s_rng = s_max - s_min

        x0 = [s_rng, 5, -s_rng / nt, s_mean]
        bounds = ([0,      0, -np.inf,      0],
                  [s_rng, nt,       0, np.inf])

        # Robust non-linear curve fit (Huber loss function)
        result = least_squares(explin, x0,
                               method='trf',
                               loss='huber',
                               bounds=bounds,
                               args=(t, s_t))

        # Fitted curve
        s_fit = explin(result.x, t, 0)

        # Add fit residual to temporal mean - detrended timeseries
        s_detrend_t[lc, :] = s_mean + result.fun

        fit_results.append(result)

    return fit_results, s_detrend_t


def explin(x, t, y):
    """
    Exponential + linear trend model

    :param x: list, parameters
        0: Exponential amplitude
        1: Exponential time constant
        2: Linear slope
        3: Offset
    :param t: array, time vector
    :param y: array, data
    :return: array, residuals
    """

    return x[0] * np.exp(-t / x[1]) + x[2] * t + x[3] - y

#     # Residual Gaussian sigma estimation
#     # Assumes Gaussian white noise + sparse outlier spikes
#     # Use robust estimate of residual sd (MAD * 1.4826)
#     phantom_res_sigma = np.median(np.abs(phantom_res)) * 1.4826
#     nyquist_res_sigma = np.median(np.abs(nyquist_res)) * 1.4826
#     noise_res_sigma = np.median(np.abs(noise_res)) * 1.4826
#
#     # Count spikes, defined as residuals more than 5 SD from zero
#     phantom_spikes = (np.abs(phantom_res) > 5 * phantom_res_sigma).sum()
#     nyquist_spikes = (np.abs(nyquist_res) > 5 * nyquist_res_sigma).sum()
#     noise_spikes = (np.abs(noise_res) > 5 * noise_res_sigma).sum()
#
#     # Calculate percent drift over course of acquisition (use fitted curve)
#     phantom_drift = (phantom_fit[nv - 1] - phantom_fit[0]) / phantom_fit[0] * 100.0
#     nyquist_drift = (nyquist_fit[nv - 1] - nyquist_fit[0]) / nyquist_fit[0] * 100.0
#     noise_drift = (noise_fit[nv - 1] - noise_fit[0]) / noise_fit[0] * 100.0
#
#     # Append mean, spike count, drift to fit parameters
#     phantom_opt = np.append(phantom_opt, [phantom_mean, phantom_spikes, phantom_drift])
#     nyquist_opt = np.append(nyquist_opt, [nyquist_mean, nyquist_spikes, nyquist_drift])
#     noise_opt = np.append(noise_opt, [noise_mean, noise_spikes, noise_drift])
#
#     # SNR relative to mean noise
#     # Estimate spatial noise sigma (assuming underlying Gaussian and Half-Normal distribution)
#     # sigma = mean(noise) * sqrt(pi/2)
#     # See for example http://en.wikipedia.org/wiki/Half-normal_distribution
#
#     noise_sigma = noise_mean * sqrt(pi / 2)
#     phantom_snr = phantom_mean / noise_sigma
#     nyquist_snr = nyquist_mean / noise_sigma
#
#     # Generate detrended timeseries - add back constant offset for each ROI
#     phantom_0 = phantom_res + phantom_mean
#     nyquist_0 = nyquist_res + nyquist_mean
#     noise_0 = noise_res + noise_mean
#
#     # Create array with detrended timeseries in columns (for output by np.savetxt)
#     ts_detrend = np.array([phantom_0, nyquist_0, noise_0])
#     ts_detrend = ts_detrend.transpose()
#
#     # Create array with timeseries detrend parameters in rows for each model
#     # Parameter order : Exp amp, Exp tau, linear, const, spike count
#     stats_pars = np.row_stack([phantom_opt, nyquist_opt, noise_opt])
#
#     #
#     # Apparent motion parameters
#     #
#     print('    Analyzing motion parameters')
#     qc_mcf_parfile = os.path.join(qc_dir, 'qc_mcf.par')
#
#     if not os.path.isfile(qc_mcf_parfile):
#         print(qc_mcf_parfile + ' does not exist - exiting')
#         sys.exit(0)
#
#     # N x 6 array (6 motion parameters in columns)
#     x = np.loadtxt(qc_mcf_parfile)
#
#     # Extract displacement timeseries for each axis
#     dx = x[:, 3]
#     dy = x[:, 4]
#     dz = x[:, 5]
#
#     # Calculate max absolute displacements (in microns) for each axis
#     max_adx = (np.abs(dx)).max() * 1000.0
#     max_ady = (np.abs(dy)).max() * 1000.0
#     max_adz = (np.abs(dz)).max() * 1000.0
#
#     #
#     # Finalize stats array for writing
#     #







# !/usr/bin/env python
"""
Quality control metric calculations from ROI timeseries fit results

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-06-04 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np


def qc_metrics(fit_results, tsfnr_nii, rois_nii):
    """
    Calculate QC metrics for each ROI

    :param fit_results: list, fit objects from scipy least_squares
    :return metrics:, dict, QC metric results dictionary
    """

    # TODO:
    # EMI analysis with zipper identification
    # Coil element SNR and fluctuation analysis
    # - Requires separate labels for each coil element

    air_mean = fit_results[0].x[3]

    nyquist_mean = fit_results[1].x[3]

    signal_a_warm = fit_results[2].x[0]
    signal_t_warm = fit_results[2].x[1]
    signal_drift = fit_results[2].x[2]
    signal_mean = fit_results[2].x[3]

    # Calculate main signal tSFNR
    tsfnr = calc_tsfnr(tsfnr_nii, rois_nii)

    # Create and fill dictionary of QC metrics
    metrics = dict()

    metrics['SignalMean'] = signal_mean
    metrics['SNR'] = signal_mean / air_mean
    metrics['SFNR'] = tsfnr
    metrics['SArtR'] = signal_mean / nyquist_mean
    metrics['Drift'] = signal_drift / signal_mean * 100
    metrics['WarmupAmp'] = signal_a_warm / signal_mean * 100
    metrics['WarmupTime'] = signal_t_warm

    # SNR relative to mean noise
    # Estimate spatial noise sigma (assuming underlying Gaussian and Half-Normal distribution)
    # sigma = mean(noise) * sqrt(pi/2)
    # See for example http://en.wikipedia.org/wiki/Half-normal_distribution

    metrics['NoiseSigma'] = air_mean * np.sqrt(np.pi/2)
    metrics['NoiseFloor'] = air_mean
    metrics['SignalSpikes'] = spike_count(fit_results[0].fun)
    metrics['NyquistSpikes'] = spike_count(fit_results[1].fun)
    metrics['AirSpikes'] = spike_count(fit_results[2].fun)

    return metrics


def spike_count(points, thresh=3.5):
    """

    :param points:
    :param thresh:
    :return:
    """

    if len(points.shape) == 1:
        points = points[:, None]

    # Median absolute deviation from the median (MAD)
    med = np.median(points, axis=0)
    dev = np.abs(points - med)
    mad = np.median(dev)

    modified_z_score = 0.6745 * dev / mad

    # Cast to int to prevent JSON encoding errors later
    return int(np.sum(modified_z_score > thresh))


def calc_tsfnr(tsfnr_nii, rois_nii):

    tsfnr_img = tsfnr_nii.get_data()
    rois_img = rois_nii.get_data()

    return np.mean(tsfnr_img[rois_img == 1])







# !/usr/bin/env python
"""
Graph plotting and image figure functions

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-05-31 JMT From scratch

This file is part of CBICQC.

   CBICQC is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   CBICQC is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
  along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.

Copyright 2019 California Institute of Technology.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from scipy.signal import periodogram
from skimage.util import montage
from skimage.exposure import rescale_intensity

from .moco import total_rotation


def plot_roi_timeseries(t, s_mean_t, s_detrend_t, plot_fname):
    """
    Plot spatial mean ROI signal vs time

    :param t: float array, time vector (s)
    :param s_mean_t:
    :param s_detrend_t:
    :param plot_fname:
    :return:
    """

    roi_names = ['Air', 'Nyquist Ghost', 'Signal']

    plt.subplots(3, 1, figsize=(10, 5))

    for lc in range(0, 3):

        plt.subplot(3, 1, lc+1)
        plt.plot(t, s_mean_t[lc, :], t, s_detrend_t[lc, :])
        plt.title(roi_names[lc], loc='left')

    # Add x label to final subplot
    plt.xlabel('Time (s)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_roi_powerspec(t, s_detrend_t, plot_fname):
    """
    Plot ROI timeseries power spectra

    :param t: float array, time vector (s)
    :param s_detrend_t:
    :param plot_fname:
    :return:
    """

    titles = ['Air (dB)', 'Nyquist Ghost (dB)', 'Signal (dB)']

    # Sampling frequency (Hz)
    fs = 1.0 / (t[1] - t[0])

    # Power spectra of detrending residual for each ROI
    f, pspec = periodogram(s_detrend_t, fs, scaling='spectrum')

    # Drop first time point (zeros)
    pspec = pspec[:, 1:]
    f = f[1:]

    plt.subplots(3, 1, figsize=(10, 5))

    for lc in range(0, 3):

        # Power dB relative to row max
        p = pspec[lc, :]
        p_max = np.max(p)
        if p_max < 1e-10:
            p_db = np.zeros_like(p)
        else:
            p_db = 10.0 * np.log10(p / np.max(p))

        plt.subplot(3, 1, lc + 1)
        plt.plot(f, p_db)
        plt.title(titles[lc], loc='left')

    # Add x label to final subplot
    plt.xlabel('Frequency (Hz)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_mopar_timeseries(t, mopars, plot_fname):
    """
    Plots x, y, z displacement and rotation timeseries from MCFLIRT registrations

    mopars columns: (dx, dy, dz, rx, ry, rz)
    Displacements in mm, rotations in degrees

    :param mopars: float array, nt x 6 motion parameters [rx, ry, rz, dx, dy, dz]
    :param plot_fname: str, output plot filename
    :return:
    """

    nt = mopars.shape[0]
    t = np.arange(0, nt)

    plt.subplots(2, 1, figsize=(10, 5))

    plt.subplot(2, 1, 1)
    plt.plot(t, mopars[:, 3:6] * 1e3)
    plt.legend(['x', 'y', 'z'])
    plt.title('Displacement (um)', loc='left')

    plt.subplot(2, 1, 2)
    plt.plot(t, mopars[:, 0:3] * 1e3)
    plt.legend(['x', 'y', 'z'])
    plt.title('Rotation (mdeg)', loc='left')

    # Add x label to final subplot
    plt.xlabel('Time (s)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def plot_mopar_powerspec(t, mopars, plot_fname):
    """
    Plot total motion power spectrum

    :param t: float array, time vector (s)
    :param mopars: array, N x 6
    :param plot_fname:
    :return:
    """

    # Sampling frequency (Hz)
    fs = 1.0 / (t[1] - t[0])

    # Total displacement timecourse
    dtot = np.linalg.norm(mopars[:, 3:6], ord=2, axis=1)

    # Total rotation timecourse
    rtot = total_rotation(mopars[:, 0:3])

    # Total motion array
    mtot = np.array([dtot, rtot])

    # Power spectra of total displacement and rotation timecourses
    f, pspec = periodogram(mtot, fs, scaling='spectrum')

    # Drop first point (zero)
    pspec = pspec[:, 1:]
    f = f[1:]

    plt.subplots(2, 1, figsize=(10, 5))
    titles = ['Displacement (dB)', 'Rotation (dB)']

    for lc in range(0, 2):

        # Power dB relative to row max
        p = pspec[lc, :]
        p_max = np.max(p)
        if p_max < 1e-10:
            p_db = np.zeros_like(p)
        else:
            p_db = 10.0 * np.log10(p / np.max(p))

        plt.subplot(2, 1, lc+1)
        plt.plot(f, p_db)
        plt.title(titles[lc], loc='left')

    # Add x axis label to last subplot
    plt.xlabel('Frequency (Hz)')

    # Space subplots without title overlap
    plt.tight_layout()

    # Save plot to file
    plt.savefig(plot_fname, dpi=300)

    # Close plot
    plt.close()


def orthoslices(img_nii, ortho_fname, cmap='viridis', irng='default'):

    orient_name = ['Axial', 'Coronal', 'Sagittal']

    img3d = img_nii.get_data()

    # Intensity scaling
    if 'robust' in irng:
        vmin, vmax = np.percentile(img3d, (1, 99))
    elif 'noscale' in irng:
        nc = plt.get_cmap(cmap).N
        vmin, vmax = 0, nc
    else:
        vmin, vmax = np.min(img3d), np.max(img3d)

    plt.subplots(1, 3, figsize=(7, 2.4))

    for ax in [0, 1, 2]:

        # Transpose dimensions for given orientation
        ax_order = np.roll([2, 0, 1], ax)
        img3d_t = np.transpose(img3d, ax_order)

        # Extract central section in first dimension
        m2d = img3d_t[int(img3d_t.shape[0]/2), :, :]

        plt.subplot(1, 3, ax+1)
        plt.imshow(m2d,
                   cmap=plt.get_cmap(cmap),
                   vmin=vmin, vmax=vmax,
                   aspect='equal',
                   origin='lower')
        plt.title(orient_name[ax])

        plt.axis('off')
        plt.subplots_adjust(bottom=0.0, top=0.9, left=0.0, right=1.0)

    # Save plot to file
    plt.savefig(ortho_fname, dpi=300)

    # Close plot
    plt.close()

    return ortho_fname


def orthoslice_montage(img_nii, montage_fname, cmap='viridis', irng='default'):

    orient_name = ['Axial', 'Coronal', 'Sagittal']

    img3d = img_nii.get_data()

    plt.subplots(1, 3, figsize=(7, 2.4))

    for ax in [0, 1, 2]:

        # Transpose dimensions for given orientation
        ax_order = np.roll([2, 0, 1], ax)
        s = np.transpose(img3d, ax_order)

        # Downsample to 9 images in first dimension
        nx = s.shape[0]
        xx = np.linspace(0, nx-1, 9).astype(int)
        s = s[xx, :, :]

        # Construct 3x3 montage of slices
        m2d = montage(s, fill='mean', grid_shape=(3, 3))

        # Intensity scaling
        if 'default' in irng:
            m2d = rescale_intensity(m2d, in_range='image', out_range=(0, 1))
        elif 'robust' in irng:
            pmin, pmax = np.percentile(m2d, (1, 99))
            m2d = rescale_intensity(m2d, in_range=(pmin, pmax), out_range=(0, 1))
        else:
            # Do nothing
            pass

        plt.subplot(1, 3, ax+1)
        plt.imshow(m2d,
                   cmap=plt.get_cmap(cmap),
                   aspect='equal',
                   origin='lower')
        plt.title(orient_name[ax])

        plt.axis('off')
        plt.subplots_adjust(bottom=0.0, top=0.9, left=0.0, right=1.0)

    # Remove excess space
    plt.tight_layout()

    # Save plot to file
    plt.savefig(montage_fname, dpi=300)

    # Close plot
    plt.close()


def roi_demeaned_ts(img_nii, rois_nii, residuals_fname):
    """
    Create temporal-spatial image of demeaned voxel timecourse
    - one graymap per ROI
    - subsample voxels in each ROI to yield n_samp timeseries

    :param img_nii:
    :param rois_nii:
    :param residuals_fname: str, ROI residuals PNG filename
    :return:
    """

    # Number of voxel samples from each ROI
    n_samp = 200

    roi_name = ['Air', 'Nyquist Ghost', 'Signal']

    rois = rois_nii.get_data()
    s = img_nii.get_data()

    # Number of time points and labels
    nt = s.shape[3]

    plt.subplots(3, 1, figsize=(7, 9))

    for lc in range(1, 4):

        mask3d = np.array(rois == lc)
        mask4d = np.tile(mask3d[:, :, :, np.newaxis], (1, 1, 1, nt))

        nx = np.sum(mask3d)

        s_xt = s[mask4d].reshape(nx, nt)

        # Downsample spatial dimension to n_samp
        inds = np.linspace(0, nx-1, n_samp).astype(int)
        s_xt_d = s_xt[inds, :]

        # Demean rows
        row_means = np.mean(s_xt_d, axis=1)
        res = s_xt_d.astype(np.float64) - np.tile(row_means[:, np.newaxis], (1, nt))

        # Plot graymap
        plt.subplot(3, 1, lc)
        plt.imshow(res,
                   cmap=plt.get_cmap('viridis'),
                   aspect='auto',
                   origin='upper'
        )

        plt.title(roi_name[lc-1])

        plt.axis('off')
        plt.subplots_adjust(bottom=0.0, top=0.9, left=0.0, right=1.0)

    # Remove excess space
    plt.tight_layout()

    # Save plot to file
    plt.savefig(residuals_fname, dpi=300)

    # Close plot
    plt.close()


def trend_subplot(m_name, t, m):
    """
    Plot session metric trend with median, 5th and 95th percentiles

    :param t: datetime, session datetime
    :param m: array, metric trend
    :return:
    """

    t0, t1 = np.min(t), np.max(t)
    p5, p50, p95 = np.percentile(m, (5, 50, 95))

    plt.gca().xaxis.set_major_locator(mdates.MonthLocator())
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter('%Y-%b'))

    plt.plot([t0, t1], [p5, p5], 'g:')
    plt.plot([t0, t1], [p50, p50], 'g')
    plt.plot([t0, t1], [p95, p95], 'g:')

    plt.plot(t, m, 'or')

    plt.gcf().autofmt_xdate()
    plt.title(m_name, loc='left')




#!/usr/bin/env python
#
# Generate an HTML calendar linking QC studies for a given scanner
#
# AUTHOR : Mike Tyszka, Ph.D.
# DATES  : 03/13/2014 JMT Adapt from trends.py
#
# This file is part of CBICQC.
#
#    CBICQC is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    CBICQC is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#   along with CBICQC.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014 California Institute of Technology.

import sys
import os
import string
import calendar
import re
import numpy as np
from matplotlib.pyplot import *
from datetime import datetime
from dateutil.relativedelta import relativedelta

# Define HTML header boilerplate
HTML_HEADER = """
<html>
<head>
<STYLE TYPE="text/css">
BODY {
  font-family    : sans-serif;
}
td {
  padding-left   : 10px;
  padding-right  : 10px;
  padding-top    : 0px;
  padding-bottom : 0px;
  vertical-align : top;
}
</STYLE>
</head>

<body>

"""

# Main function
def main():

  # Get scanner name from arguments
  if len(sys.argv) > 1:
    scanner_name = sys.argv[1]
  else:
    print('USAGE : calendar.py <scanner name>')
    sys.exit(1)

  # Get QC data directory from shell environment
  cbicqc_data_dir = os.environ['CBICQC_DATA']

  # Full scanner QC directory path
  scanner_qc_dir = os.path.join(cbicqc_data_dir, scanner_name)

  # Output calendar web page name
  # NOTE : this is the index page for the scanner QC directory
  qc_scanner_calendar = os.path.join(scanner_qc_dir, 'index.html')

  print('Writing scanner calendar to ' + qc_scanner_calendar)

  # Set first day of week to SUNDAY (US)
  calendar.setfirstweekday(calendar.SUNDAY)

  #
  # HTML report generation
  #

  # Open HTML calendar page for this scanner
  fd = open(qc_scanner_calendar, "w")

  # Write HTML header boilerplate
  fd.write(HTML_HEADER)

  # Write header splash for this scanner
  fd.write('<h1 style="background-color:#E0E0FF">CBIC QC Calendar for %s</h1>' %(scanner_name))

  # Create a table to hold month calendars
  # Two rows of three months, from oldest to most recent

  fd.write('<table><tr>')

  # Note today's datetime
  dt_today = datetime.today()

  # Loop over the past twelve months, including current month
  for dmonth in range(-5, 1, 1):
  
    # Create a new row at month -2
    if dmonth == -2:
      fd.write('</tr><tr>')

    # Get the datetime object for dmonths ago
    dt_past = dt_today + relativedelta( months = dmonth)

    # Start new table element
    fd.write('<td>')
    
    # Write table header for this month
    fd.write('<table>')
    fd.write('<tr><td colspan="7"><center><b>%s %d</b></center></tr>' %(dt_past.strftime("%B"), dt_past.year))
    fd.write('<tr><td>Sun<td>Mon<td>Tue<td>Wed<td>Thu<td>Fri<td>Sat</tr>')

    # Create month calendar array
    past_cal = calendar.monthcalendar(dt_past.year, dt_past.month)
    
    # Number of weeks in this month
    nwk = len(past_cal)

    # Loop over weeks
    for wk in range(0, nwk):
    
      # New table row
      fd.write('<tr align="center">')
    
      # Look over days in week
      for d in range(0,7):
      
        this_d = past_cal[wk][d]
        
        if this_d > 0:
        
          # Create QC date string (YYYYMMDD)
          qc_date_str = '%s%02d%02d' %(dt_past.year, dt_past.month, this_d)
          
          # Create path to local daily QC index page
          qc_dir = os.path.join(scanner_qc_dir, qc_date_str)
          qc_index = os.path.join(qc_dir, 'index.html')
          
          # Check if daily QC index exists for this date
          if os.path.isfile(qc_index):
          
            # Create relative URL to this daily QC
            # No need to prefix with scanner name
            qc_url = os.path.join(qc_date_str, 'index.html')
            fd.write('<td><a href="%s">%d</a>' %(qc_url, this_d))
      
          else:
            
            # Just write plain date without link
            fd.write('<td>%d' %(this_d))
      
    
        else:
        
          fd.write('<td>')

      # Finish row for this week
      fd.write('</tr>')

    # Finish table for this month
    fd.write('</table><p>')

  # Finish table of month calendars
  fd.write('</table>')

#---------------------------------------------

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()



# !/usr/bin/env python
"""
QC analysis node for nipype

AUTHOR : Mike Tyszka
PLACE  : Caltech
DATES  : 2019-05-22 JMT From scratch

MIT License

Copyright (c) 2019 Mike Tyszka

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import subprocess
import pkg_resources
import nibabel as nb
import numpy as np
from scipy.ndimage.morphology import (binary_dilation,
                                      binary_erosion,
                                      generate_binary_structure)


def register_template(tmean_nii, work_dir, mode='phantom'):
    """
    Register the appropriate template to the provided tmean image
    - for phantom QC, a translation-only registration is performed
    - for in vivo QC, an affine registration is performed

    :param tmean_nii: Nifti object,
        Temporal mean image
    :param work_dir: str,
        Full path to working directory
    :param mode: str,
        QC mode, 'phantom' or 'live'
    :return:
    """

    # Save temporal mean image for use by FLIRT
    tmean_fname = os.path.join(work_dir, 'fixed.nii.gz')
    nb.save(tmean_nii, tmean_fname)

    # Link appropriate template for mode
    if 'phantom' in mode:
        dof = 6
        template_fname = pkg_resources.resource_filename(
            __name__,
            os.path.join('templates', 'fbirn_sphere.nii.gz')
        )
        labels_fname = pkg_resources.resource_filename(
            __name__,
            os.path.join('templates', 'fbirn_labels.nii.gz')
        )
    else:
        dof = 12
        template_fname = pkg_resources.resource_filename(
            __name__,
            os.path.join('templates', 'MNI152_T1_2mm.nii.gz')
        )
        labels_fname = pkg_resources.resource_filename(
            __name__,
            os.path.join('templates', 'MNI-maxprob-thr25-2mm.nii.gz')
        )

    template_xfm_fname = os.path.join(work_dir, 'template_xfm.nii.gz')
    labels_xfm_fname = os.path.join(work_dir, 'labels_xfm.nii.gz')

    fsl_dir = os.environ['FSLDIR']
    flirt_cmd = os.path.join(fsl_dir, 'bin', 'flirt')
    xfm_fname = os.path.join(work_dir, 'xfm.mat')

    # Run FLIRT registration
    print('      Registering template to subject ({} DOF)'.format(dof))
    cmd = [flirt_cmd,
           '-in', template_fname,
           '-ref', tmean_fname,
           '-out', template_xfm_fname,
           '-dof', str(dof),
           '-omat', xfm_fname,
    ]
    subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)

    # Apply resulting transform to label image
    print('      Resampling labels to subject space')
    cmd = [flirt_cmd,
           '-in', labels_fname,
           '-ref', tmean_fname,
           '-out', labels_xfm_fname,
           '-applyxfm',
           '-init', xfm_fname,
           '-interp', 'nearestneighbour',
    ]
    subprocess.run(cmd, stderr=sys.stderr, stdout=sys.stdout)

    # Load labels transformed to subject space
    print('      Loading resampled labels')
    labels_nii = nb.load(labels_xfm_fname)

    return labels_nii


def make_rois(labels_nii):
    """
    Organize labels and add air space and Nyquist ROIs

    :param labels_nii: Nifti object,
        Raw labels in subject space
    :return rois_nii: Nifti object,
        Integer ROI image include air and Nyquist ghost
    """

    # Extract label image
    labels_img = labels_nii.get_data().astype(np.uint)

    # Create signal mask from sum of all ROI labels
    signal_mask = labels_img > 0

    # Construct 3D binary sphere structuring element
    k = generate_binary_structure(3, 2)  # 3D, 2-connected

    # Erode signal mask x N, then dilate x 2N
    n_iter = 2
    signal_mask_ero = binary_erosion(signal_mask, structure=k, iterations=n_iter)
    signal_mask_dil = binary_dilation(signal_mask_ero, structure=k, iterations=n_iter * 2)

    # Create Nyquist mask by rolling eroded signal mask by FOVy/2
    ny = signal_mask.shape[1]
    nyquist_mask = np.roll(signal_mask_ero, int(ny / 2), axis=1)

    # Exclusive Nyquist ghost mask (remove overlap with dilated signal mask)
    nyquist_only_mask = np.logical_and(nyquist_mask, np.logical_not(np.logical_and(nyquist_mask, signal_mask_dil))).astype(np.uint)

    # Create air mask
    air_mask = (1 - signal_mask_dil - nyquist_only_mask).astype(np.uint)

    # Finally merge all masks into an ROI label file
    # Undefined       = 0
    # Air Space       = 1
    # Nyquist Ghost   = 2
    # Other ROIs      = 3, 4, 5, ...
    rois = air_mask + 2 * nyquist_only_mask + labels_img + 2 * signal_mask

    # Wrap image in a Nifti object
    rois_nii = nb.Nifti1Image(rois, labels_nii.affine)

    return rois_nii
